#Jan 21'19 - converted to work on Ruby and adding nogoods as the whole answer set showing up
#Jan 24'19 - also works on Schubert
#          - takes as first input the interested atoms &  only looking at positive interested atoms in the answer sets to add in nogoods
#          - limited the number of ans sets to 50
# Jan 26'19 - only add one atom at a time as nogood
#!/usr/bin/env python3

import clingo
import os, sys, random

class Thing:
    def __init__(self):
        pass

def maxnum(list):
    max=0
    for i in range(0,len(list)):
      temp=list[i][1].number
      if temp>max:
            max=temp
    return max

class ModelReceiver:
  def __init__(self, sample_count, sample_atoms, interesting_atoms, cannot_atoms):
    self.displayed = 0
    self.displayed_best = 0
    self.currentbest = None
    self.lastalltrue = None
    self.sample_count = sample_count
    self.sample_atoms = sample_atoms
    self.interesting_atoms = interesting_atoms
    self.cannot_atoms = cannot_atoms

  def __call__(self, m):
            # suppress duplicate answer sets
            alltrue = m.symbols(atoms=True)
            if alltrue == self.lastalltrue:
              # skip answer set
              return
            else:
              self.lastalltrue = alltrue

            # collect information about this answer set
            here = { 'cost': [], 'violated': [], 'interesting': [], 'valid': '  valid' }
            here['cost'] = m.cost
            if self.currentbest != m.cost:
              if self.currentbest:
                print("IGNORE RESULTS ABOVE (COST HAS CHANGED)")
              self.currentbest = m.cost
              self.displayed_best = 0
            isTrue = lambda x: m.is_true(x.literal)
            for atom in filter(isTrue, self.cannot_atoms):
                here['valid'] = 'invalid'
                for elem in filter(isTrue, atom.elements):
                    here['violated'].append(elem.message)
            here['interesting'] = [ x.symbol for x in self.interesting_atoms if m.is_true(x.literal) ]
            
            self.displayed += 1
            self.displayed_best += 1
            print("ANS {0[cost]} {0[valid]}".format(here))
            for what in ['interesting']:
              print("    {:12s} {}".format(what.upper(), repr(here[what])))
            for v in here['violated']:
              print("    {:12s} {}".format('VIOLATED', v))

            if self.sample_atoms > 0:
              print(here['interesting'])  
              true_non_facts = [x for x in here['interesting'] if not m.context.symbolic_atoms[x].is_fact]
              for s in range(0, self.sample_count):
                if len(true_non_facts)>self.sample_atoms:
                  to_forbid = random.sample(true_non_facts, self.sample_atoms)
                else:  
                  to_forbid=true_non_facts  
                #to_forbid=true_non_facts
                print("adding nogood {} for more diversity".format(repr([str(x) for x in to_forbid])))
                m.context.add_nogood([ (x, True) for x in to_forbid ])



def main():
    print clingo.__version__
    print("arguments: {}".format(repr(sys.argv[1:])))

    SAMPLE_ATOMS = 1 # how many to forbid in each nogood
    SAMPLE_COUNT = 1 # how many nogoods to create
    if SAMPLE_ATOMS > 0:
      print("Mode: sampling {} random constraints over {} atoms".format(SAMPLE_COUNT, SAMPLE_ATOMS))
    else:
      print("Mode: finding all answer sets")
    
    interested=sys.argv[1].split(",")
    # split into asp files and other arguments
    aspfiles = [ x for x in sys.argv[2:] if ".lp" in x or ".txt" in x]#x.endswith('.lp') ]
    clingoargs = [ x for x in sys.argv[2:] if ".lp" not in x and ".txt" not in x]#not x.endswith('.lp') ]

    # --quiet=2,2,2
    prg = clingo.Control(clingoargs)

    # load all files given in path
    for aspfile in aspfiles:
      prg.load(aspfile)

    # add cannot theory
    prg.add('base', [], """
      #theory cannot {
          t { };
          &cannot/0: t, head
      }.
      """)

    # ground
    prg.ground([("base", [])])

    # extract relevant theory atoms and minimize the number of messages
    #with prg.backend() as backend:
    backend=prg.backend	
    lits = []
    lits_time = []
    cannot_atoms = []
    for atom in prg.theory_atoms:
        #print("atom {}".format(repr(atom)))
        if atom.term.name == "cannot" and len(atom.term.arguments) == 0:
            a = Thing()
            a.literal = atom.literal
            a.elements = []
            for elem in atom.elements:
                e = Thing()
                
		# format message
                t = elem.terms
                e.message = t[0].name[1:-1]
                time_point = t[len(t)-1]
                #print t
                for term in t[1:]:
                    e.message = e.message.replace("%", str(term), 1)

                # create condition to minimize
                condition = elem.condition[:]
                condition.append(atom.literal)
                if len(condition) > 1:
                    atom = backend.add_atom()
                    backend.add_rule([atom], condition)
                    condition[0] = atom
                lits.append((condition[0], 1))
                lits_time.append((condition[0], time_point))
                e.literal = condition[0]

                a.elements.append(e)
            cannot_atoms.append(a)

    interesting_atoms = []
    for atom in prg.symbolic_atoms:
        #print("atom {} has program literal {}".format(atom.symbol, atom.literal))
        # Jan 24' 19 : only looking at positive ones
        if atom.symbol.name in interested and atom.symbol.positive:#['go']:
            a = Thing()
            a.literal = atom.literal
            a.symbol = atom.symbol
            interesting_atoms.append(a)
    #print lits_time
    if len(lits_time)>0:
        #max_time=max(lits_time,key=itemgetter(1))[1].number
        max_time=maxnum(lits_time)
        #print max_time
        lits_time_rearrange=[]
        for t in range(0,max_time+1):
          lits_time_rearrange.append([])
          lits_time_rearrange[t].append([(item[0],t) for item in lits_time if item[1].number == t])
          backend.add_minimize(-t, lits_time_rearrange[t][0])
          #print lits_time_rearrange

    receiver = ModelReceiver(SAMPLE_COUNT, SAMPLE_ATOMS, interesting_atoms, cannot_atoms)

    # get all optimal models
    prg.configuration.solve.models = 50
    prg.configuration.solve.opt_mode = 'optN'

    prg.solve(on_model=receiver)
    print("displayed {} answer sets ({} optimal answer sets)".format(receiver.displayed, receiver.displayed_best))

if __name__ == '__main__':
  main()
